@@ -10,11 +10,12 @@ import uuid
 import warnings
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Optional
+from typing import Any, Callable, Dict, List, Optional, Union
 
 import requests
 
 from annofabapi import AnnofabApi
+from annofabapi.api import my_backoff
 from annofabapi.exceptions import AnnofabApiException
 from annofabapi.models import (
     AdditionalData,
@@ -28,7 +29,6 @@ from annofabapi.models import (
     Inspection,
     InspectionStatus,
     Instruction,
-    JobInfo,
     JobStatus,
     JobType,
     LabelV1,
@@ -36,6 +36,8 @@ from annofabapi.models import (
     Organization,
     OrganizationMember,
     Project,
+    ProjectJobInfo,
+    ProjectJobType,
     ProjectMember,
     SimpleAnnotationDetail,
     SupplementaryData,
@@ -80,35 +82,50 @@ _ORGANIZATION_ID_FOR_AVAILABILITY = "___plannedWorktime___"
 """予定稼働時間用の組織ID"""
 
 _JOB_CONCURRENCY_LIMIT = {
-    JobType.COPY_PROJECT: {JobType.GEN_INPUTS, JobType.GEN_TASKS, JobType.DELETE_PROJECT, JobType.MOVE_PROJECT},
-    JobType.GEN_INPUTS: {
-        JobType.COPY_PROJECT,
-        JobType.GEN_INPUTS,
-        JobType.GEN_TASKS,
-        JobType.GEN_INPUTS_LIST,
-        JobType.DELETE_PROJECT,
-        JobType.MOVE_PROJECT,
+    ProjectJobType.COPY_PROJECT: {
+        ProjectJobType.GEN_INPUTS,
+        ProjectJobType.GEN_TASKS,
+        ProjectJobType.DELETE_PROJECT,
+        ProjectJobType.MOVE_PROJECT,
     },
-    JobType.GEN_TASKS: {
-        JobType.COPY_PROJECT,
-        JobType.GEN_INPUTS,
-        JobType.GEN_TASKS,
-        JobType.GEN_ANNOTATION,
-        JobType.GEN_TASKS_LIST,
-        JobType.DELETE_PROJECT,
-        JobType.MOVE_PROJECT,
+    ProjectJobType.GEN_INPUTS: {
+        ProjectJobType.COPY_PROJECT,
+        ProjectJobType.GEN_INPUTS,
+        ProjectJobType.GEN_TASKS,
+        ProjectJobType.GEN_INPUTS_LIST,
+        ProjectJobType.DELETE_PROJECT,
+        ProjectJobType.MOVE_PROJECT,
     },
-    JobType.GEN_ANNOTATION: {JobType.GEN_TASKS, JobType.GEN_ANNOTATION, JobType.DELETE_PROJECT, JobType.MOVE_PROJECT},
-    JobType.GEN_TASKS_LIST: {JobType.GEN_TASKS, JobType.GEN_TASKS_LIST, JobType.DELETE_PROJECT, JobType.MOVE_PROJECT},
-    JobType.GEN_INPUTS_LIST: {
-        JobType.GEN_INPUTS,
-        JobType.GEN_INPUTS_LIST,
-        JobType.DELETE_PROJECT,
-        JobType.MOVE_PROJECT,
+    ProjectJobType.GEN_TASKS: {
+        ProjectJobType.COPY_PROJECT,
+        ProjectJobType.GEN_INPUTS,
+        ProjectJobType.GEN_TASKS,
+        ProjectJobType.GEN_ANNOTATION,
+        ProjectJobType.GEN_TASKS_LIST,
+        ProjectJobType.DELETE_PROJECT,
+        ProjectJobType.MOVE_PROJECT,
     },
-    JobType.INVOKE_HOOK: {JobType.DELETE_PROJECT, JobType.MOVE_PROJECT},
-    JobType.DELETE_PROJECT: set(JobType),
-    JobType.MOVE_PROJECT: set(JobType),
+    ProjectJobType.GEN_ANNOTATION: {
+        ProjectJobType.GEN_TASKS,
+        ProjectJobType.GEN_ANNOTATION,
+        ProjectJobType.DELETE_PROJECT,
+        ProjectJobType.MOVE_PROJECT,
+    },
+    ProjectJobType.GEN_TASKS_LIST: {
+        ProjectJobType.GEN_TASKS,
+        ProjectJobType.GEN_TASKS_LIST,
+        ProjectJobType.DELETE_PROJECT,
+        ProjectJobType.MOVE_PROJECT,
+    },
+    ProjectJobType.GEN_INPUTS_LIST: {
+        ProjectJobType.GEN_INPUTS,
+        ProjectJobType.GEN_INPUTS_LIST,
+        ProjectJobType.DELETE_PROJECT,
+        ProjectJobType.MOVE_PROJECT,
+    },
+    ProjectJobType.INVOKE_HOOK: {ProjectJobType.DELETE_PROJECT, ProjectJobType.MOVE_PROJECT},
+    ProjectJobType.DELETE_PROJECT: set(ProjectJobType),
+    ProjectJobType.MOVE_PROJECT: set(ProjectJobType),
 }
 """同時に実行できないジョブを表しています。valueに指定されたジョブが1つ以上実行されている場合、keyに指定されたジョブは実行できません。"""
 
@@ -201,24 +218,19 @@ class Wrapper:
     #########################################
     # Public Method : Annotation
     #########################################
-    def download_annotation_archive(self, project_id: str, dest_path: str, v2: bool = False) -> str:
+    def download_annotation_archive(self, project_id: str, dest_path: str) -> str:
         """
         simpleアノテーションZIPをダウンロードする。
 
         Args:
             project_id: プロジェクトID
             dest_path: ダウンロード先のファイルパス
-            v2:  互換性を保つために残している引数。2020-10-31以降に廃止する。
 
         Returns:
             ダウンロード元のURL
 
         """
         query_params = None
-        if v2:
-            query_params = {"v2": True}
-            logger.warning("メソッド引数`v2`は2020-10-31以降に廃止します。")
-
         _, response = self.api.get_annotation_archive(project_id, query_params=query_params)
         url = response.headers["Location"]
         response2 = _download(url, dest_path)
@@ -239,7 +251,7 @@ class Wrapper:
             ダウンロード元のURL
 
         """
-        warnings.warn("deprecated", DeprecationWarning)
+        warnings.warn("deprecated", FutureWarning)
         _, response = self.api.get_archive_full_with_pro_id(project_id)
         url = response.headers["Location"]
         response2 = _download(url, dest_path)
@@ -263,14 +275,6 @@ class Wrapper:
             self.api.get_annotation_list, limit=200, project_id=project_id, query_params=query_params
         )
 
-    @staticmethod
-    def __create_annotation_id(detail: Dict[str, Any]) -> str:
-        if detail["data_holding_type"] == AnnotationDataHoldingType.INNER.value and detail["data"] is None:
-            # annotation_typeがclassificationのときは、label_idとannotation_idを一致させる必要がある。
-            return detail["label_id"]
-        else:
-            return str(uuid.uuid4())
-
     @staticmethod
     def __replace_annotation_specs_id(
         detail: Dict[str, Any], annotation_specs_relation: AnnotationSpecsRelation
@@ -649,6 +653,9 @@ class Wrapper:
         """
         アノテーション仕様を、別のプロジェクトにコピーする。
 
+        .. deprecated:: 2021-09-01
+
+
         Note:
             誤って実行しないようにすること
 
@@ -660,6 +667,8 @@ class Wrapper:
         Returns:
             put_annotation_specsのContent
         """
+        warnings.warn("2021-09-01以降に削除します。", FutureWarning)
+
         src_annotation_specs = self.api.get_annotation_specs(src_project_id)[0]
 
         if comment is None:
@@ -906,7 +915,8 @@ class Wrapper:
     #########################################
     # Public Method : Statistics
     #########################################
-    def _request_location_header_url(self, response: requests.Response) -> Any:
+    @my_backoff
+    def _request_location_header_url(self, response: requests.Response) -> Optional[Any]:
         """
         Location headerに記載されているURLの中身を返す。
 
@@ -914,10 +924,15 @@ class Wrapper:
             response:
 
         Returns:
-            Location headerに記載されているURLの中身
+            Location headerに記載されているURLの中身。
+            レスポンスヘッダにLocationがない場合は、Noneを返す。
 
         """
-        url = response.headers["Location"]
+        url = response.headers.get("Location")
+        if url is None:
+            # プロジェクト作成直後などが該当する
+            logger.warning(f"レスポンスヘッダに'Location'がありません。method={response.request.method}, url={response.request.url}")
+            return None
 
         response = self.api.session.get(url)
         _log_error_response(logger, response)
@@ -939,7 +954,11 @@ class Wrapper:
 
         """
         _, response = self.api.get_task_statistics(project_id)
-        return self._request_location_header_url(response)
+        result = self._request_location_header_url(response)
+        if result is not None:
+            return result
+        else:
+            return []
 
     def get_account_statistics(self, project_id: str) -> List[Any]:
         """
@@ -952,7 +971,11 @@ class Wrapper:
 
         """
         _, response = self.api.get_account_statistics(project_id)
-        return self._request_location_header_url(response)
+        result = self._request_location_header_url(response)
+        if result is not None:
+            return result
+        else:
+            return []
 
     def get_inspection_statistics(self, project_id: str) -> List[Any]:
         """
@@ -965,7 +988,11 @@ class Wrapper:
 
         """
         _, response = self.api.get_inspection_statistics(project_id)
-        return self._request_location_header_url(response)
+        result = self._request_location_header_url(response)
+        if result is not None:
+            return result
+        else:
+            return []
 
     def get_task_phase_statistics(self, project_id: str) -> List[Any]:
         """
@@ -978,7 +1005,11 @@ class Wrapper:
 
         """
         _, response = self.api.get_task_phase_statistics(project_id)
-        return self._request_location_header_url(response)
+        result = self._request_location_header_url(response)
+        if result is not None:
+            return result
+        else:
+            return []
 
     def get_label_statistics(self, project_id: str) -> List[Any]:
         """
@@ -991,7 +1022,11 @@ class Wrapper:
 
         """
         _, response = self.api.get_label_statistics(project_id)
-        return self._request_location_header_url(response)
+        result = self._request_location_header_url(response)
+        if result is not None:
+            return result
+        else:
+            return []
 
     def get_worktime_statistics(self, project_id: str) -> List[Any]:
         """
@@ -1006,7 +1041,11 @@ class Wrapper:
 
         """
         _, response = self.api.get_worktime_statistics(project_id)
-        return self._request_location_header_url(response)
+        result = self._request_location_header_url(response)
+        if result is not None:
+            return result
+        else:
+            return []
 
     #########################################
     # Public Method : Supplementary
@@ -1345,152 +1384,6 @@ class Wrapper:
         content, _ = self.api.get_project_members(project_id, query_params=query_params)
         return content["list"]
 
-    def put_project_members(self, project_id, project_members: List[Dict[str, Any]]) -> List[ProjectMember]:
-        """
-        複数のプロジェクトメンバを追加/更新/削除する.
-
-        .. deprecated:: 2020-01-01
-
-        Args:
-            project_id: プロジェクトID
-            project_members: 追加/更新するメンバのList. `user_id` , `member_status` , `member_role` をKeyに持つこと
-
-        Returns:
-            `putProjectMember` APIのContentのList
-
-        """
-        warnings.warn("deprecated: 2021-01-01以降に廃止します。", DeprecationWarning)
-
-        # 追加/更新前のプロジェクトメンバ
-        dest_project_members = self.get_all_project_members(project_id)
-
-        updated_project_members = []
-        # プロジェクトメンバを追加/更新する
-        for member in project_members:
-            dest_member = [e for e in dest_project_members if e["user_id"] == member["user_id"]]
-            last_updated_datetime = dest_member[0]["updated_datetime"] if len(dest_member) > 0 else None
-
-            request_body = {
-                "member_status": member["member_status"],
-                "member_role": member["member_role"],
-                "sampling_inspection_rate": member.get("sampling_inspection_rate"),
-                "sampling_acceptance_rate": member.get("sampling_acceptance_rate"),
-                "last_updated_datetime": last_updated_datetime,
-            }
-            updated_project_member = self.api.put_project_member(
-                project_id, member["user_id"], request_body=request_body
-            )[0]
-            updated_project_members.append(updated_project_member)
-
-            command_name = "追加" if last_updated_datetime is None else "更新"
-            logger.debug(
-                "プロジェクトメンバの'%s' 完了." " project_id=%s, user_id=%s, " "last_updated_datetime=%s",
-                command_name,
-                project_id,
-                member["user_id"],
-                last_updated_datetime,
-            )
-
-        return updated_project_members
-
-    def assign_role_to_project_members(
-        self, project_id: str, user_id_list: List[str], member_role: str
-    ) -> List[ProjectMember]:
-        """
-        複数のプロジェクトメンバに1つのロールを割り当てる。
-
-        .. deprecated:: 2020-01-01
-
-        Note:
-            誤って実行しないようにすること
-
-        Args:
-            project_id: プロジェクトID
-            user_id_list: 追加/更新するメンバのuser_idのList
-            member_role: 割り当てるロール.
-
-        Returns:
-            `putProjectMember` APIのContentのList
-
-        """
-        warnings.warn("deprecated: 2021-01-01以降に廃止します。", DeprecationWarning)
-
-        project_members = []
-        for user_id in user_id_list:
-            member = {"user_id": user_id, "member_status": "active", "member_role": member_role}
-            project_members.append(member)
-
-        return self.put_project_members(project_id, project_members)
-
-    def drop_role_to_project_members(self, project_id, user_id_list: List[str]) -> List[ProjectMember]:
-        """
-        複数のプロジェクトメンバを、プロジェクトから脱退させる
-
-        .. deprecated:: 2020-01-01
-
-        Note:
-            誤って実行しないようにすること
-
-        Args:
-            project_id: プロジェクトID
-            user_id_list: 脱退させるメンバのuser_idのList
-
-        Returns:
-            `putProjectMember` APIのContentのList
-        """
-        warnings.warn("deprecated: 2021-01-01以降に廃止します。", DeprecationWarning)
-
-        project_members = []
-        for user_id in user_id_list:
-            member = {
-                "user_id": user_id,
-                "member_status": "inactive",
-                "member_role": "worker",  # 何か指定しないとエラーになったため、指定する
-            }
-            project_members.append(member)
-
-        return self.put_project_members(project_id, project_members)
-
-    def copy_project_members(
-        self, src_project_id: str, dest_project_id: str, delete_dest: bool = False
-    ) -> List[ProjectMember]:
-        """
-        プロジェクトメンバを、別のプロジェクトにコピーする。
-
-        .. deprecated:: 2020-01-01
-
-        Note:
-            誤って実行しないようにすること
-
-        Args:
-            src_project_id: コピー元のproject_id
-            dest_project_id: コピー先のproject_id
-            delete_dest: Trueならばコピー先にしか存在しないプロジェクトメンバを削除する。
-
-        Returns:
-            `putProjectMember` APIのContentのList
-
-        """
-        warnings.warn("deprecated: 2021-01-01以降に廃止します。", DeprecationWarning)
-
-        src_project_members = self.get_all_project_members(src_project_id)
-        dest_project_members = self.get_all_project_members(dest_project_id)
-
-        if delete_dest:
-            # コピー先にしかいないメンバを削除する
-            src_account_ids = [e["account_id"] for e in src_project_members]
-            deleted_dest_members = [e for e in dest_project_members if e["account_id"] not in src_account_ids]
-
-            def to_inactive(arg_member):
-                arg_member["member_status"] = "inactive"
-                return arg_member
-
-            deleted_dest_members = list(map(to_inactive, deleted_dest_members))
-            return self.put_project_members(dest_project_id, src_project_members + deleted_dest_members)
-
-        else:
-            return self.put_project_members(dest_project_id, src_project_members)
-
     #########################################
     # Public Method : Task
     #########################################
@@ -1834,7 +1727,9 @@ class Wrapper:
     #########################################
     # Public Method : Job
     #########################################
-    def delete_all_succeeded_job(self, project_id: str, job_type: JobType) -> List[JobInfo]:
+    def delete_all_succeeded_job(
+        self, project_id: str, job_type: Union[ProjectJobType, JobType]
+    ) -> List[ProjectJobInfo]:
         """
         成功したジョブをすべて削除する
 
@@ -1845,7 +1740,6 @@ class Wrapper:
         Returns:
             削除したジョブの一覧
         """
-
         jobs = self.get_all_project_job(project_id, {"type": job_type.value})
         deleted_jobs = []
         for job in jobs:
@@ -1856,7 +1750,7 @@ class Wrapper:
 
         return deleted_jobs
 
-    def get_all_project_job(self, project_id: str, query_params: Dict[str, Any]) -> List[JobInfo]:
+    def get_all_project_job(self, project_id: str, query_params: Dict[str, Any] = None) -> List[ProjectJobInfo]:
         """
         すべてのバックグランドジョブを取得する。
 
@@ -1877,7 +1771,7 @@ class Wrapper:
         all_jobs.extend(r["list"])
         return all_jobs
 
-    def job_in_progress(self, project_id: str, job_type: JobType) -> bool:
+    def job_in_progress(self, project_id: str, job_type: Union[ProjectJobType, JobType]) -> bool:
         """
         ジョブが進行中かどうか
 
@@ -1897,7 +1791,11 @@ class Wrapper:
         return job["job_status"] == JobStatus.PROGRESS.value
 
     def wait_for_completion(
-        self, project_id: str, job_type: JobType, job_access_interval: int = 60, max_job_access: int = 10
+        self,
+        project_id: str,
+        job_type: Union[ProjectJobType, JobType],
+        job_access_interval: int = 60,
+        max_job_access: int = 10,
     ) -> bool:
         """
         ジョブが完了するまで待つ。
@@ -1925,7 +1823,7 @@ class Wrapper:
     def wait_until_job_finished(
         self,
         project_id: str,
-        job_type: JobType,
+        job_type: Union[ProjectJobType, JobType],
         job_id: Optional[str] = None,
         job_access_interval: int = 60,
         max_job_access: int = 360,
@@ -1946,14 +1844,14 @@ class Wrapper:
 
         """
 
-        def get_latest_job() -> Optional[JobInfo]:
+        def get_latest_job() -> Optional[ProjectJobInfo]:
             job_list = self.api.get_project_job(project_id, query_params={"type": job_type.value})[0]["list"]
             if len(job_list) > 0:
                 return job_list[0]
             else:
                 return None
 
-        def get_job_from_job_id(arg_job_id: str) -> Optional[JobInfo]:
+        def get_job_from_job_id(arg_job_id: str) -> Optional[ProjectJobInfo]:
             content, _ = self.api.get_project_job(project_id, query_params={"type": job_type.value})
             job_list = content["list"]
             return _first_true(job_list, pred=lambda e: e["job_id"] == arg_job_id)
@@ -2003,19 +1901,24 @@ class Wrapper:
                     )
                     return JobStatus.PROGRESS
 
-    async def _job_in_progress_async(self, project_id: str, job_type: JobType) -> bool:
+    async def _job_in_progress_async(self, project_id: str, job_type: Union[ProjectJobType, JobType]) -> bool:
         loop = asyncio.get_event_loop()
         return await loop.run_in_executor(None, self.job_in_progress, project_id, job_type)
 
     async def _wait_until_job_finished_async(
-        self, project_id: str, job_type: JobType, job_id: Optional[str], job_access_interval: int, max_job_access: int
+        self,
+        project_id: str,
+        job_type: Union[ProjectJobType, JobType],
+        job_id: Optional[str],
+        job_access_interval: int,
+        max_job_access: int,
     ) -> Optional[JobStatus]:
         loop = asyncio.get_event_loop()
         return await loop.run_in_executor(
             None, self.wait_until_job_finished, project_id, job_type, job_id, job_access_interval, max_job_access
         )
 
-    def can_execute_job(self, project_id: str, job_type: JobType) -> bool:
+    def can_execute_job(self, project_id: str, job_type: Union[ProjectJobType, JobType]) -> bool:
         """
         ジョブが実行できる状態か否か。他のジョブが実行中で同時に実行できない場合はFalseを返す。
 
@@ -2026,21 +1929,30 @@ class Wrapper:
         Returns:
             ジョブが実行できる状態か否か
         """
-        job_type_list = _JOB_CONCURRENCY_LIMIT[job_type]
+        # TODO: JobTypeが削除されたら、この処理も削除する
+        new_job_type: ProjectJobType = ProjectJobType(job_type.value) if isinstance(job_type, JobType) else job_type
+
+        job_type_list = _JOB_CONCURRENCY_LIMIT[new_job_type]
 
         # tokenがない場合、ログインが複数回発生するので、事前にログインしておく
         if self.api.token_dict is None:
             self.api.login()
 
         # 複数のジョブに対して進行中かどうかを確認する
-        gather = asyncio.gather(*[self._job_in_progress_async(project_id, job_type) for job_type in job_type_list])
+        gather = asyncio.gather(
+            *[self._job_in_progress_async(project_id, new_job_type) for new_job_type in job_type_list]
+        )
         loop = asyncio.get_event_loop()
         result = loop.run_until_complete(gather)
 
         return all(not e for e in result)
 
     def wait_until_job_is_executable(
-        self, project_id: str, job_type: JobType, job_access_interval: int = 60, max_job_access: int = 360
+        self,
+        project_id: str,
+        job_type: Union[ProjectJobType, JobType],
+        job_access_interval: int = 60,
+        max_job_access: int = 360,
     ) -> bool:
         """
         ジョブが実行可能な状態になるまで待ちます。他のジョブが実行されているときは、他のジョブが終了するまで待ちます。
@@ -2055,7 +1967,10 @@ class Wrapper:
             指定した時間（アクセス頻度と回数）待った後、ジョブが実行可能な状態かどうか。進行中のジョブが存在する場合は、ジョブが実行不可能。
 
         """
-        job_type_list = _JOB_CONCURRENCY_LIMIT[job_type]
+        # TODO: JobTypeが削除されたら、この処理も削除する
+        new_job_type: ProjectJobType = ProjectJobType(job_type.value) if isinstance(job_type, JobType) else job_type
+
+        job_type_list = _JOB_CONCURRENCY_LIMIT[new_job_type]
         # tokenがない場合、ログインが複数回発生するので、事前にログインしておく
         if self.api.token_dict is None:
             self.api.login()
@@ -2063,8 +1978,8 @@ class Wrapper:
         # 複数のジョブに対して進行中かどうかを確認する
         gather = asyncio.gather(
             *[
-                self._wait_until_job_finished_async(project_id, job_type, None, job_access_interval, max_job_access)
-                for job_type in job_type_list
+                self._wait_until_job_finished_async(project_id, new_job_type, None, job_access_interval, max_job_access)
+                for new_job_type in job_type_list
             ],
             return_exceptions=True,
         )
